/* File: Plotstream.h
 *
 * Class Plotstream
 * A plotstream opens a window, displays a data plot, then closes
 * the window when the user presses a key.
 *
 * Author: 	jlk
 * Version:	1.1
 * Date:	July 2005
 */
#pragma once

#include "PlotData.h"

enum Rounding{DOWN,ANY,UP};

template<class T> bool setMin(T&min,T v) {if (min>v) {min=v; return true;} return false;}
template<class T> bool setMax(T&max,T v) {if (max<v) {max=v; return true;} return false;}

template<class T>class satint{
 T x;
 enum klasse{	// muss unsigned ausgewertet werden!!
  PINF,		// 0 = +Inf (7FFF)
  NNAN,		// 1 = NaN (8000)
  NINF,		// 2 = -Inf (8001)
 };
 static klasse test(T v) {return klasse(v-std::numeric_limits<T>::max());}
 static T untest(klasse v) {return v+std::numeric_limits<T>::max();}
 T add(T y) const{
  klasse kx=test(x);
  klasse ky=test(y);
  if (kx>=3 && ky>=3) {
   T r=x+y;
// TODO: Check overflow flag! Machine-dependent!
   return r;
  }
  if (kx>=3) swap(kx,ky);	// kx ist bevorzugt "ungltig"
  if (kx==NNAN || ky==NNAN) return untest(NNAN);
  if (kx+ky==2) return untest(NNAN);	// zwei entgegengesetzte Unendlich = NaN
  return untest(kx);		// Unendlich von einem der beiden Operanden
 }
 T mul(T y) const{
  klasse kx=test(x);
  klasse ky=test(y);
  if (kx>=3 && ky>=3) {
   T r=x*y;
// TODO: Check overflow flag! Machine-dependent!
   return r;
  }
  if (kx>=3) {swap(kx,ky); y=x;}
  if (kx==NNAN || ky==NNAN) return untest(NNAN);
  if (y==0) return untest(NNAN);	// Unendlich mal Null = NaN
  if (kx==PINF && y<0) return untest(NINF);
  if (kx==NINF && y>=0) return untest(NINF);
  return untest(PINF);
 }
 T div(T y) const{
  klasse kx=test(x);
  klasse ky=test(y);
  if (kx>=3 && ky>=3) {
   T r=x/y;
// TODO: Check overflow flag! Machine-dependent!
   return r;
  }
 }
public:
// unr
 T operator-() const{return -x;}
 T operator+() const{return x;}
 T operator--() {return x=add(-1);}
 T operator--(int) {T r=x; x=add(-1); return r;}
 T operator++() {return x=add(1);}
 T operator++(int) {T r=x; x=add(1); return r;}
// binr
 T operator+(T v) const{return add(v);}
 T operator-(T v) const{return add(-v);}
 T operator<<(int n) const{return shl(n);}
 T operator>>(int n) const{return shl(-n);}
 T operator*(T v) const{return mul(v);}
 T operator/(T v) const{return div(v);}
 bool operator<(T v) const{T r=add(-v); return klasse(r)!=NNAN && r<0;}
 bool operator<=(T v) const{T r=add(-v); return klasse(r)!=NNAN && r<=0;}
 bool operator==(T v) const{return !add(-v);}
 bool operator!=(T v) const{return !operator==(v);}
 bool operator>(T v) const{return add(-v)>0;}
 bool operator>=(T v) const{return add(-v)>=0;}
// Bitoperationen (&,|,~,^) sind mit satint nicht erlaubt!
// Zuweisung
 T operator=(T v) {return x=v;}
 T operator+=(T v) {return x=add(v);}
 T operator-=(T v) {return x=add(-v);}
 T operator*=(T v) {return x=mul(v);}
 T operator/=(T v) {return x=div(v);}
 T operator<<=(int n) {return x=shl(n);}
 T operator>>=(int n) {return x=shl(-n);}
 template<class TF> operator TF() const	// in float oder double konvertieren, gerne auch satint<short> in satint<long>
	{switch (test(x)) {
	  case PINF: return std::numeric_limits<TF>::infinity();
	  case NNAN: return std::numeric_limits<TF>::quiet_NaN();
	  case NINF: return -std::numeric_limits<TF>::infinity();
	  default: return TF(x);
	 }
	}
// satint(T v):x(v) {}
 template<class TF> satint(TF v)	// von float oder double erstellen, gerne auch von satint<long> oder als Kopierkonstruktor wenn T==TF
	{if (v!=v) x=untest(NNAN);
	 else if (v>untest(PINF)) x=untest(PINF);
	 else if (v<untest(NINF)) x=untest(NINF);
	 else x=T(v);
	}
};
// Ist die Klasse so sinnvoll gebaut? Was fehlt?
// Wie kann ich std::numeric_limits<> auf satint<T> erweitern?

template<class T=double> class Plotstream{
public:
// Plotstream(const char * title = "2D Plot", int width = 640, int height = 0);
 Plotstream(const char*title=0);
 ~Plotstream();
 void addplot(const Plotdata<T>&x, const Plotdata<T>&y, Color colour = GREEN);
 void show(const char*title=0);
 void plot(const Plotdata<T>&x, const Plotdata<T>&y, Color colour = GREEN);
 static HWND wnd;
 static HDC dc;
 void onPaint();
 struct attrib{
  Color colour;
  char penwidth;
  char penstyle;
  char fillto;
  char drawstyle;
 };
 struct xytrace{
  const Plotdata<T>*x;
  const Plotdata<T>*y;
  attrib a;
 };
private:
 Rect<int> rcPlot;
// int winWidth, winHeight;
// int plotWidth, plotHeight;
// char * winTitle;
 Range<T> rx,ry;	// Highest and lowest data values in x and y
 LinScale<T,int> sx,sy;	// Scales of graph drawing to screen pixels
 bool plotStarted;	// True while plotting is going on
 bool marked; 		// True when a marker is visible
 int lastX;
 int lastY; 		// Location of last marker drawn
 Color colour;	// Current drawing colour
 Color lastColour;	// Previous drawing colour
	   
	/* Convert graph x value to screen coordinate */
 int X(T x) const	{return sx.scale(x);}
	/* Convert graph y value to screen coordinate */
 int Y(T y) const	{return sy.scale(y);}
	/* Convert screen coordinate to graph x value */
 T plotX(int screenX) const	{return sx.unscale(screenX);}
	/* Convert screen coordinate to graph y value */
 T plotY(int screenY) const	{return sy.unscale(screenY);}
	/* Round(est) double within one pixel in X */
 Rounding nearRoundX(T & val, Rounding direction = ANY) const;
	/* Round(est) double within one pixel in Y */
 Rounding nearRoundY(T & val, Rounding direction = ANY) const;
	   /* True if values given are within x and y graph ranges */
 bool withinRange(T xVal, T yVal) const;
	/* Draws the axes */
 void drawAxes();
	/* Watches the mouse and prints cursor coords if clicked */
// bool watchMouse();
	// Draw a marker at the current cursor position. Erase if erase is true
 void drawMarker(bool erase = false);
    // Draw a single point at the requested coordinates
 void drawSinglePoint(T xCoord, T yCoord) const;
    // Set new foreground colour, store last colour
// void setFgColor(Color fgColour);
    // Reset foreground colour to last colour
// void resetFgColor();
    // Check whether the stream holds a command at this point (colour change
    // or single point), and handle the command if it does.
// void handleCommand(dataIterator&x_it,dataIterator&y_it);
 
 void drawPointShape(int x, int y) const;
 void drawMarkShape(int x, int y) const;
 bool lineStarted;
 void lineStart() {lineStarted=false;}
 void lineto(int x, int y) const {if (lineStarted) LineTo(dc,x,y); else MoveToEx(dc,x,y,0),lineStarted=true;}
 void outtextxy(int x, int y,const char*s) const {TextOut(dc,x,y,s,strlen(s));}
 struct gdiobj{
  HPEN penPlot;
  HBRUSH brFill;
 };
 struct marker_t{
  T x,y;
 };
 struct internal_xytrace{
  xytrace t;
  std::vector<marker_t>markers;
  gdiobj g;
 };
 std::vector<internal_xytrace> traces;
	/* Draw the data */
 void drawFunc(internal_xytrace&t);
};
